# This module load the outputs saved for each individual forecast system
# and creates the multimodel ensemble

source("modules/Loading/R/dates2load.R")
source("modules/Loading/R/get_timeidx.R")
source("modules/Loading/R/check_latlon.R")
source('modules/Multimodel/load_multimodel.R')

Multimodel <- function(recipe) {
  
  # recipe: auto-s2s recipe as provided by read_yaml
  
  # Loading data saved in the jobs for individual models
  if (tolower(recipe$Analysis$Datasets$Multimodel$split_loading) %in% c('true','yes')){
    
    ## TO DO: 
    #Create new function load_multimodel.R, until fcst, with startR retrieve false
    #Load data with retrieve = F to check ensemble dimensions of each system
    #Create array of NA with dimension of maximum ensemble members
    #Loop over systems adding data into created array with abind
    #Add dimension names since lost with abind
    #Add filled array into data object
    
    # Retrieve data dimension only without loading data
    source('modules/Multimodel/load_multimodel_split.R')
    dims <- load_multimodel_split(recipe)    
    
    # Create empty array with desired dimensions
    hcst_aux <- array(data = NA, dim = dims$dim.hcst)
    
    if (!is.null(recipe$Analysis$Time$fcst_year)) {
      fcst_aux <- array(data = NA, dim = dims$dim.fcst)
    } else {
      fcst_aux <- NULL
    }
    
    data_order <- names(dims$dim.hcst)
    
    # ## Loop over model to load data 
    for (sys in 1:length(recipe$Analysis$Datasets$System$models)){
      
      system_load <- recipe$Analysis$Datasets$System$models[sys]

      recipe_aux <- recipe
      recipe_aux$Analysis$Datasets$System$models <- system_load

      data_aux <- load_multimodel(recipe_aux)
      
      hcst_aux[,,,,,which(recipe$Analysis$Datasets$System$models == system_load),,,,] <- s2dv::Reorder(data = data_aux$hcst$data, order = data_order)

      ## Error:  Not able to combine arrays with different 'ensemble' dimensions
      
    }
     
    ## Necessary to do separately for hcst and fcst since can different ensemble dimensions
    
    # data$hcst$data <- hcst_aux
    
    # if(!is.null(recipe$Analysis$Time$fcst_year)){
    #   data$fcst$data <- fcst_aux     
    # } else {
    # data$fcst$data <- NULL

  } ## close if on split_loading
  
    ## Load data without splitting
    data <- load_multimodel(recipe)    
  }
  
  # Creating the multi-model
  if (tolower(recipe$Analysis$Datasets$Multimodel$approach) == 'pooled') {
    
    # Deterministic hindcast
    data$hcst$data <- CSTools::MergeDims(data = data$hcst$data, 
                                         merge_dims = c('model','ensemble'), 
                                         rename_dim = 'ensemble', na.rm = TRUE)
    data$hcst$dims <- dim(data$hcst$data)
    
    # Deterministic forecast
    if (!is.null(recipe$Analysis$Time$fcst_year)) {
      data$fcst$data <- CSTools::MergeDims(data = data$fcst$data, 
                                           merge_dims = c('model','ensemble'), 
                                           rename_dim = 'ensemble', na.rm = TRUE)
      data$fcst$dims <- dim(data$fcst$data)
    }
    
    # Saving multimodel
    if (recipe$Analysis$Workflow[[recipe$Analysis$Datasets$Multimodel$createFrom]]$save != 'none') {
      save_forecast(recipe = recipe,
                    data_cube = data$hcst,
                    outdir = outdir[var],
                    type = 'hcst')
      if (!is.null(data$fcst)) {
        save_forecast(recipe = recipe, 
                      data_cube = data$fcst,
                      outdir = outdir[var],
                      type = 'fcst')
      }
      save_observations(recipe = recipe, 
                        data_cube = data$obs,
                        outdir = outdir[var])
    }
    
    # Probabilistic hindcast and forecast
    prob <- Probabilities(recipe, data)
    
  } else if (tolower(recipe$Analysis$Datasets$Multimodel$approach) %in% c('mean','median')) {
    
    # Probabilistic hindcast and forecast
    warning('Probabilities for multi-model mean is still under development.
             If Skill is used, the results are not correct for the probabilitic metrics.
             Probabilities cannot be computed because data has an extra dimension (model).
             Also, the function should return the observed probabilities.
             Maybe it is better to use GetProbs.')
    prob <- NULL #Probabilities(recipe, data)
    
    # Deterministic hindcast
    data$hcst$data <- multiApply::Apply(data = data$hcst$data, 
                                        target_dims = 'ensemble',
                                        fun = tolower(recipe$Analysis$Datasets$Multimodel$approach),
                                        na.rm = TRUE,
                                        ncores = recipe$Analysis$ncores)$output1
    data$hcst$data <- multiApply::Apply(data = data$hcst$data,
                                        target_dims = 'model',
                                        fun = tolower(recipe$Analysis$Datasets$Multimodel$approach),
                                        na.rm = FALSE,
                                        ncores = recipe$Analysis$ncores)$output1
    data$hcst$dims <- s2dv::InsertDim(data = data$hcst$dims, posdim = 6, lendim = 1, name = 'ensemble')
    data$hcst$dims <- dim(data$hcst$data)
    
    # Deterministic forecast
    if (!is.null(recipe$Analysis$Time$fcst_year)) {
      data$fcst$data <- multiApply::Apply(data = data$fcst$data, 
                                          target_dims = 'ensemble',
                                          fun = tolower(recipe$Analysis$Datasets$Multimodel$approach),
                                          na.rm = TRUE,
                                          ncores = recipe$Analysis$ncores)$output1
      data$fcst$data <- multiApply::Apply(data = data$fcst$data,
                                          target_dims = 'model',
                                          fun = tolower(recipe$Analysis$Datasets$Multimodel$approach),
                                          na.rm = FALSE,
                                          ncores = recipe$Analysis$ncores)$output1
      data$fcst$dims <- s2dv::InsertDim(data = data$fcst$dims, posdim = 6, lendim = 1, name = 'ensemble')
      data$fcst$dims <- dim(data$fcst$data)
    }
    
  } else {stop('Incorrect multi-model approach')}
  
  # Removing the temporary data of models in case user only requests multimodel
  if (tolower(recipe$Analysis$Datasets$Multimodel$execute) %in% c('true','yes')) {
    
    if (tolower(recipe$Analysis$Horizon) == 'seasonal') {
      exp.name <- recipe$Analysis$Datasets$System$models
    } else if (tolower(recipe$Analysis$Horizon) == 'decadal') {
      exp.name <- sapply(recipe$Analysis$Datasets$System$models, '[[', 'name')
    } else {stop('Multimodel not implemented for this horizon.')}
    unlink(file.path(recipe$Run$output_dir,"outputs/*/*",gsub('\\.','',exp.name)),
           recursive = T)
    unlink(file.path(recipe$Run$output_dir,"plots/*",gsub('\\.','',exp.name)),
           recursive = T)
    
  } else if (tolower(recipe$Analysis$Datasets$Multimodel$execute) == 'both' &&
             recipe$Analysis$Workflow[[recipe$Analysis$Datasets$Multimodel$createFrom]]$save == 'none') {
    
    if (tolower(recipe$Analysis$Horizon) == 'seasonal') {
      exp.name <- recipe$Analysis$Datasets$System$models
    } else if (tolower(recipe$Analysis$Horizon) == 'decadal') {
      exp.name <- sapply(recipe$Analysis$Datasets$System$models, '[[', 'name')
    } else {stop('Multimodel not implemented for this horizon.')}
    exp.name <- c(exp.name,'Multimodel')
    unlink(file.path(recipe$Run$output_dir,"outputs/",
                     recipe$Analysis$Datasets$Multimodel$createFrom),
           recursive = T)
    
  }
  
  return(list(data = data, prob = prob))
}
